<?php

namespace FiloBlu\Refilo\Remote\Entity\Provider;

use FiloBlu\Refilo\Remote\Entity\AbstractEntity;
use FiloBlu\Refilo\Remote\Entity\EntityProviderInterface;
use FiloBlu\Refilo\Remote\Entity\Product;
use Magento\CatalogInventory\Model\Stock;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\DB\Select;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Module\Manager;
use Magento\Framework\ObjectManagerInterface;
use Magento\InventorySalesApi\Api\Data\SalesChannelInterface;
use Magento\InventorySalesApi\Api\GetStockBySalesChannelInterface;
use Magento\Store\Api\Data\StoreInterface;
use Magento\Store\Api\Data\WebsiteInterface;
use Magento\Store\Model\StoreManagerInterface;
use Zend_Db;

/**
 * Class StockProvider
 * @package FiloBlu\Refilo\Remote\Entity\Provider
 */
class StockProvider extends BaseProvider
{
	/** @var int */
	protected $website;

	/** @var Manager */
	protected $moduleManager;

	/** @var ResourceConnection */
	private $resourceConnection;

	/** @var ObjectManagerInterface */
	private $objectManager;

	/** @var StoreManagerInterface */
	private $storeManager;

	/**
	 * StockProvider constructor.
	 * @param Manager $moduleManager
	 * @param ObjectManagerInterface $objectManager
	 * @param ResourceConnection $resourceConnection
	 * @param StoreManagerInterface $storeManager
	 * @param ScopeConfigInterface $scopeConfig
	 */
	public function __construct(
		Manager                $moduleManager,
		ObjectManagerInterface $objectManager,
		ResourceConnection     $resourceConnection,
		StoreManagerInterface  $storeManager,
		ScopeConfigInterface   $scopeConfig
	)
	{
		parent::__construct($scopeConfig);
		$this->resourceConnection = $resourceConnection;
		$this->moduleManager = $moduleManager;
		$this->objectManager = $objectManager;
		$this->storeManager = $storeManager;
	}

	/**
	 *
	 * @throws NoSuchEntityException|LocalizedException
	 */
        #[\ReturnTypeWillChange]
	public function rewind()
	{
		$this->items = [];
		$stockId = Stock::DEFAULT_STOCK_ID;

		if ($this->moduleManager->isEnabled('Magento_Inventory') && $this->website) {
			/** @var GetStockBySalesChannelInterface $getStockBySalesChannel */
			$getStockBySalesChannel = $this->objectManager->get(GetStockBySalesChannelInterface::class);
			/** @var  SalesChannelInterface $salesChannel */
			$salesChannel = $this->objectManager->create(SalesChannelInterface::class);
			$salesChannel->setType(SalesChannelInterface::TYPE_WEBSITE);
			$salesChannel->setCode($this->storeManager->getWebsite($this->website)->getCode());
			$stock = $getStockBySalesChannel->execute($salesChannel);
			$stockId = $stock->getStockId();
		}

		$select = $this->getSelectForStock($stockId);

		if (!empty($this->ids)) {
			$select->where('cpe.entity_id IN (' . implode(',', $this->ids) . ')');
		}

		foreach ($this->resourceConnection->getConnection()->fetchAll($select, [], Zend_Db::FETCH_ASSOC) as $item) {
			$productId = (int)$item['id'];
			$this->items[$productId] = new Product([
				AbstractEntity::ID => $productId,
				Product::INVENTORY => [
					'in_stock'   => $item['is_salable'] > 0,
					'backorders' => (bool)$item['backorders'],
					'saleable'   => $item['is_salable'] > 0
				]
			]);
		}

		if (!$this->hasReadHandler()) {
			return;
		}

		// TODO: Handle pagination
		// Temporary workaround for Morato
		$pageSize = $this->getBulkSize();
		$size = count($this->items);

		while ($size > $pageSize) {
			$data = array_slice($this->items, 0, $pageSize);
			$this->getReadHandler()->onRead($data, $pageSize, $this->getReadHandlerArguments());
			$size -= $pageSize;
		}

		if ($size > 0) {
			$this->getReadHandler()->onRead($this->items, $size, $this->getReadHandlerArguments());
		}
	}

	/**
	 * @param $stockId
	 * @return Select
	 */
	protected function getSelectForStock($stockId)
	{
		$usePreorderAttribute = (boolean)$this->scopeConfig->getValue('sync/data/use_preorder_attribute');
		return $this->resourceConnection->getConnection()
			->select()
			->from(
				[
					's' => "inventory_stock_{$stockId}"
				],
				[
					'id'         => 'cpe.entity_id',
					'quantity'   => 'quantity',
					'is_salable' => 'is_salable',
					'backorders' => ($usePreorderAttribute) ? 'IFNULL(csi.value,0)' : 'IFNULL(csi.backorders,0)'
				]
			)->join(
				[
					'cpe' => 'catalog_product_entity'
				],
				's.sku = cpe.sku',
				[]
			)->joinLeft(
				['csi' => ($usePreorderAttribute) ? 'catalog_product_entity_int' : 'cataloginventory_stock_item'],
				($usePreorderAttribute) ? 'cpe.row_id = csi.row_id AND csi.attribute_id = (select attribute_id from eav_attribute WHERE attribute_code = \'is_preorder\') AND csi.store_id = 0' : 'cpe.entity_id = csi.product_id',
				[]
			);
	}

	/**
	 * @param array $ids
	 * @return $this|EntityProviderInterface
	 */
	public function withIds(array $ids): EntityProviderInterface
	{
		$this->ids = $ids;
		return $this;
	}

	/**
	 * @param StoreInterface $store
	 * @return $this|EntityProviderInterface
	 */
	public function forStore(StoreInterface $store): EntityProviderInterface
	{
		$this->website = $store->getWebsiteId();
		return $this;
	}

	/**
	 * @param WebsiteInterface $website
	 * @return $this|EntityProviderInterface
	 */
	public function forWebsite(WebsiteInterface $website): EntityProviderInterface
	{
		$this->website = $website->getId();
		return $this;
	}
}
